
'''
This file is automatically generated; Do not edit it. :)
'''
VERSION_INFO = {
    'final': True,
    'version': '38.4.0',
    'branch_nick': 'release-38-LTS',
    'revision_id': '945ab71be90243fe96148bb3bd13c1c3ae410765',
    'revno': 14030
}
